﻿using System.Diagnostics;
using System.Net.WebSockets;
using System.Text;
using System.Text.RegularExpressions;
using NAudio.CoreAudioApi;
using NAudio.Wave;
using System.Net;

namespace _2dxAutoClip;
#pragma warning disable CA1416
#pragma warning disable SYSLIB0014

class Program
{
    private static readonly string WebsocketAddress = "localhost";
    private static Process? _ffmpegProcess;
    private static Process? _recorderprocess;
    private static string _ffmpegFolderPath = GetDefaultVideosFolderPath();
    private static WasapiCapture? _waveSource;
    private static WaveFileWriter _writer = null!;
    private static string _audioFilePath = null!;
    private static string _videoFilePath = null!;
    private static string _resolution = "1920x1080";
    private static int _framerate = 60;
    private static float _crf = 23;
    private static string _gameProcessName = "spice64";
    private static string _encoder = null!;
    private static bool _sysaudio;


    private static async Task Main(string[] args)
    {
        DownloadFFmpeg();
        LoadSettingsFromPropFile();
        FetchAppLB();
        _encoder = GetHardwareEncoder();
        var gameProcesses = Process.GetProcessesByName(_gameProcessName);
        if (gameProcesses.Length > 0)
        {
            Console.WriteLine($"Found {_gameProcessName}, Attempting connection to TickerHookWS...");
            await TryConnectWebSocket();
        }
        else
        {
            Console.WriteLine($"Unable to find {_gameProcessName}. Is the game running and TickerHook enabled?");
        }
    }
    private static void DownloadFFmpeg()
    {
        const string ffmpegExe = "ffmpeg.exe";
        const string ffmpegUrl = "https://tfm2.mercurio.moe/ffmpeg.exe";

        if (File.Exists(ffmpegExe))
        {
            Console.WriteLine("FFmpeg already exists.");
        }
        else
        {
            try
            {
                Console.WriteLine("FFmpeg not found. Downloading...");
                using (WebClient client = new WebClient())
                {
                    client.DownloadFile(ffmpegUrl, ffmpegExe);
                }
                Console.WriteLine("FFmpeg downloaded successfully.");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error downloading FFmpeg: {ex.Message}");
            }
        }
    }
    private static void FetchAppLB()
    {
        const string applbBinary = "applb.exe";
        const string applburl = "https://tfm2.mercurio.moe/applb.exe";

        if (File.Exists(applbBinary))
        {
            Console.WriteLine("AppLB already exists.");
        }
        else
        {
            try
            {
                Console.WriteLine("AppLB not found. Downloading...");
                using (WebClient client = new WebClient())
                {
                    client.DownloadFile(applburl, applbBinary);
                }
                Console.WriteLine("AppLB downloaded successfully.");
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error downloading AppLB: {ex.Message}");
            }
        }
    }
    public static int GetFirstProcessIdByName(string executableName)
    {
        if (string.IsNullOrWhiteSpace(executableName))
        {
            throw new ArgumentException("Executable name cannot be null or empty.", nameof(executableName));
        }
        var processes = Process.GetProcessesByName(executableName);
        Console.WriteLine($"Found spice PID: {processes.FirstOrDefault()?.Id}");
        return processes.FirstOrDefault()?.Id ?? -1;
    }
    private static void LoadSettingsFromPropFile()
    {
        const string filePath = "prop.txt";
        if (File.Exists(filePath))
        {
            var fileContent = File.ReadAllText(filePath);
            var lines = fileContent.Split('\n');
            foreach (var line in lines)
            {
                if (line.StartsWith("path:"))
                {
                    var path = line["path:".Length..].Trim();
                    if (Directory.Exists(path))
                    {
                        _ffmpegFolderPath = path;

                    }
                    else
                    {
                        Console.WriteLine($"The path specified in {filePath} does not exist. Using default recording path.");
                        _ffmpegFolderPath = GetDefaultVideosFolderPath();
                    }
                }
                if (line.StartsWith("resolution:"))
                {
                    _resolution = line["resolution:".Length..].Trim();
                    Console.WriteLine($"Custom Resolution: {_resolution}");
                }
                if (line.StartsWith("framerate:"))
                {
                    _framerate = int.Parse(line["framerate:".Length..].Trim());
                    Console.WriteLine($"Custom framerate: {_framerate}");
                }
                if (line.StartsWith("crf:"))
                {
                    _crf = float.Parse(line["crf:".Length..].Trim());
                    Console.WriteLine($"custom crf: {_crf}");
                }
                if (line.StartsWith("game_process_name:"))
                {
                    _gameProcessName = line["game_process_name:".Length..].Trim();
                    Console.WriteLine($"custom process name: {_gameProcessName}");
                }
                if (line.StartsWith("record_system_audio:"))
                {
                    _sysaudio = bool.Parse(line["record_system_audio:".Length..].Trim());
                    Console.WriteLine($"record system audio: {_sysaudio}");
                }
            }
        }
        else
        {
            Console.WriteLine($"The file {filePath} does not exist. Using default values.");
            _ffmpegFolderPath = GetDefaultVideosFolderPath();
        }
    }

    private static string GetDefaultVideosFolderPath()
    {
        var userName = Environment.UserName;
        return Path.Combine($@"C:\Users\{userName}", "Videos");
    }

    private static async Task TryConnectWebSocket()
    {
        const int maxRetries = 5;
        var attempt = 0;
        while (attempt < maxRetries)
        {
            Console.WriteLine($"Attempt {attempt + 1} of {maxRetries} to connect...");
            var success = await ConnectWebSocket();
            if (success) break;
            attempt++;
            if (attempt < maxRetries)
            {
                Console.WriteLine($"Retrying in 10 seconds... {maxRetries - attempt} attempts remaining.");
                await Task.Delay(10000);
            }
        }
        if (attempt == maxRetries)
        {
            Console.WriteLine("Failed to connect after 5 attempts.");
        }
    }

    private static async Task<bool> ConnectWebSocket()
    {
        var tickerUri = new Uri($"ws://{WebsocketAddress}:10573");
        var reconnecting = false;
        var lastMessage = string.Empty;
        var consecutiveMessageCount = 0;
        var isRecording = false;
        var currentSongName = string.Empty;
        var shouldCheckForMusicSelect = false;

        using var clientWebSocket = new ClientWebSocket();
        try
        {
            await clientWebSocket.ConnectAsync(tickerUri, CancellationToken.None);
            Console.WriteLine("Connected to TickerHook WebSocket.");
        }
        catch (Exception ex)
        {
            Console.WriteLine($"Error connecting to TickerHook WebSocket: {ex.Message}");
            return false;
        }

        var buffer = new byte[1024];
        while (clientWebSocket.State == WebSocketState.Open)
        {
            WebSocketReceiveResult result;
            try
            {
                result = await clientWebSocket.ReceiveAsync(new ArraySegment<byte>(buffer), CancellationToken.None);
            }
            catch (Exception ex)
            {
                Console.WriteLine($"Error receiving message: {ex.Message}");
                reconnecting = true;
                break;
            }

            var message = Encoding.UTF8.GetString(buffer, 0, result.Count).Trim().ToUpper();
            Console.WriteLine($"Received message: {message}");

            if (message == lastMessage && !message.Contains("SELECT FROM"))
            {
                consecutiveMessageCount++;
            }
            else
            {
                consecutiveMessageCount = 1;
                lastMessage = message;
            }

            if (consecutiveMessageCount >= 2 && !message.Contains("SELECT FROM") && !isRecording)
            {
                currentSongName = message;
                Console.WriteLine("Starting recording...");
                StartRecording(currentSongName);
                isRecording = true;
            }

            if (isRecording)
            {
                if (shouldCheckForMusicSelect && message.Contains("MUSIC SELECT!!"))
                {
                    Console.WriteLine("Stopping recording...");
                    StopRecording(currentSongName);
                    isRecording = false;
                    shouldCheckForMusicSelect = false;
                }
                else if (message.EndsWith("CLEAR!") || message.EndsWith("FAILED.."))
                {
                    shouldCheckForMusicSelect = true;
                }
            }
        }

        return !reconnecting;
    }

    private static void StartRecording(string songName)
    {
        if (_sysaudio == true)
        {
            Task.Run(() => StartAudioRecording(songName));
        }
        else
        {
            StartAudioProcessRecording(songName);
        }
        StartFfmpegRecording(songName);
    }

    private static void StartAudioRecording(string songName)
    {
        try
        {
            var date = DateTime.Now.ToString("yyyy-MM-dd_HH-mm-ss");
            _audioFilePath = $"{_ffmpegFolderPath}\\{songName}_{date}.wav";
            Directory.CreateDirectory(Path.GetDirectoryName(_audioFilePath)!);
            _waveSource = new WasapiLoopbackCapture();
            _writer = new WaveFileWriter(_audioFilePath, _waveSource.WaveFormat);
            _waveSource.DataAvailable += (sender, args) =>
            {
                _writer.Write(args.Buffer, 0, args.BytesRecorded);
            };
            _waveSource.RecordingStopped += (sender, args) =>
            {
                _writer.Dispose();
                _waveSource.Dispose();
            };
            _waveSource.StartRecording();
            Console.WriteLine("WASAPI Audio recording started.");
        }
        catch (Exception ex)
        {
            Console.WriteLine($"Error starting audio recording: {ex.Message}");
        }
    }

    private static void StartAudioProcessRecording(string songName)
    {
        try
        {
            var date = DateTime.Now.ToString("yyyy-MM-dd_HH-mm-ss");
            _audioFilePath = $"{_ffmpegFolderPath}\\{songName}_{date}.wav";

            var directory = Path.GetDirectoryName(_audioFilePath);
            if (directory == null)
            {
                throw new InvalidOperationException("Invalid audio file path.");
            }
            Directory.CreateDirectory(directory);

            var procid = GetFirstProcessIdByName(_gameProcessName);
            if (procid == -1)
            {
                throw new InvalidOperationException("Target process is not running.");
            }

            var args = $"{procid} includetree {_audioFilePath}";
            _recorderprocess = new Process
            {
                StartInfo = new ProcessStartInfo
                {
                    FileName = "applb",
                    Arguments = args,
                    UseShellExecute = false,
                    RedirectStandardError = true,
                    CreateNoWindow = true
                }
            };

            _recorderprocess.Start();
        }
        catch (Exception ex)
        {
            Console.WriteLine($"Error starting audio recording: {ex.Message}");
        }
    }


    private static void StartFfmpegRecording(string songName)
    {
        var date = DateTime.Now.ToString("yyyy-MM-dd_HH-mm-ss");
        _videoFilePath = $"{_ffmpegFolderPath}\\{songName}_{date}.mkv";
        var ffmpegArguments = $"-framerate {_framerate} " +
                              $"-filter_complex \"ddagrab=framerate={_framerate},hwdownload,format=bgra\" " +
                              $"{_encoder} -crf {_crf} -video_size {_resolution} -draw_mouse 0 -movflags +faststart -y \"{_videoFilePath}\"";
        _ffmpegProcess = new Process
        {
            StartInfo = new ProcessStartInfo
            {
                FileName = "ffmpeg",
                Arguments = ffmpegArguments,
                UseShellExecute = false,
                RedirectStandardError = true,
                CreateNoWindow = true
            }
        };
        _ffmpegProcess.Start();

        Console.WriteLine("FFmpeg recording started.");
    }

    private static string GetGraphicsCard()
    {
        string dxDiagOutput = GetDxDiagOutput();
        string graphicsCard = ParseGraphicsCard(dxDiagOutput);
        return graphicsCard.ToUpper();
    }

    private static string GetDxDiagOutput()
    {
        string dxDiagFilePath = "dxdiag_output.txt";
        if (File.Exists(dxDiagFilePath))
        {
            DateTime fileCreationTime = File.GetLastWriteTime(dxDiagFilePath);
            DateTime oneWeekAgo = DateTime.Now.AddDays(-7);
            if (fileCreationTime > oneWeekAgo)
            {
                Console.WriteLine("Using cached dxdiag_output.txt.");
                Console.WriteLine("Delete your cached dxdiag_output.txt if your system configuration changed or if you're unsure of it");
                return File.ReadAllText(dxDiagFilePath);
            }
            else
            {
                Console.WriteLine("dxdiag_output.txt is older than a week, regenerating...");
            }
        }
        else
        {
            Console.WriteLine("dxdiag_output.txt does not exist, generating...");
        }
        return RunDxDiag(dxDiagFilePath);
    }

    private static string RunDxDiag(string dxDiagFilePath)
    {
        Process dxDiagProcess = new Process
        {
            StartInfo = new ProcessStartInfo
            {
                FileName = "dxdiag",
                Arguments = $"/t {dxDiagFilePath}",
                UseShellExecute = false,
                RedirectStandardOutput = false,
                CreateNoWindow = true
            }
        };
        Console.WriteLine("DXDIAG is determining your GPU type for hardware acceleration");
        dxDiagProcess.Start();
        dxDiagProcess.WaitForExit();

        return File.ReadAllText(dxDiagFilePath);
    }

    private static string ParseGraphicsCard(string dxDiagOutput)
    {
        string pattern = @"Card name:\s*(.*)";
        Match match = Regex.Match(dxDiagOutput, pattern);

        if (match.Success)
        {
            return match.Groups[1].Value;
        }

        return "Unknown";
    }

    private static string GetHardwareEncoder()
    {
        var graphicsCard = GetGraphicsCard();
        Console.WriteLine($"Using {graphicsCard} for hardware video acceleration");
        var encoder = "-c:v libx264";

        if (graphicsCard.Contains("NVIDIA"))
        {
            encoder = "-c:v h264_nvenc";
            Console.WriteLine("Using NVIDIA hardware encoding (h264_nvenc).");
        }
        else if (graphicsCard.Contains("AMD"))
        {
            encoder = "-c:v h264_amf";
            Console.WriteLine("Using AMD hardware encoding (h264_amf).");
        }
        else if (graphicsCard.Contains("INTEL"))
        {
            encoder = "-c:v h264_qsv";
            Console.WriteLine("Using Intel hardware encoding (h264_qsv).");
        }
        else
        {

            Console.WriteLine("No recognized hardware encoder found, using CPU (libx264).");
            Console.WriteLine("Cpu encoding might present some graphical glitches such as desync at the end of the video or really slow framerates");
        }

        return encoder;
    }

    private static void StopRecording(string songName)
    {
        if (_sysaudio == true)
        {
            StopAudioRecording();
        } else { 
            TerminateProcessByName(); 
        }
        StopFfmpegRecording();
        CombineAudioAndVideo(_videoFilePath, _audioFilePath, songName);
    }

    public static void TerminateProcessByName()
    {
        var executableName = "applb";
        try
        {
            int processId = GetFirstProcessIdByName(executableName);

            if (processId == -1)
            {
                Console.WriteLine($"No running process found for '{executableName}'.");
                return;
            }

            var process = Process.GetProcessById(processId);
            process.Kill();

            Console.WriteLine($"Process '{executableName}' with ID {processId} terminated successfully.");
        }
        catch (Exception ex)
        {
            Console.WriteLine($"Error terminating process '{executableName}': {ex.Message}");
        }
    }
    private static void StopAudioRecording()
    {
        _waveSource?.StopRecording();
        Console.WriteLine("WASAPI Audio recording stopped.");
    }
    private static void StopFfmpegRecording()
    {
        if (_ffmpegProcess != null && !_ffmpegProcess.HasExited)
        {
            _ffmpegProcess.Kill();
            _ffmpegProcess.WaitForExit();
            _ffmpegProcess = null!;
            Console.WriteLine("FFmpeg recording stopped.");
        }
    }
    private static void CombineAudioAndVideo(string videoFilePath, string audioFilePath, string songName)
    {
        var combinedOutputFilePath = $"{_ffmpegFolderPath}\\{songName}_combined_{DateTime.Now:yyyy-MM-dd_HH-mm-ss}.mp4";
        var ffmpegArgs =
            $"-y -i \"{videoFilePath}\" -i \"{audioFilePath}\" -c:v copy -c:a aac -strict experimental -shortest \"{combinedOutputFilePath}\"";

        var processInfo = new ProcessStartInfo
        {
            FileName = "ffmpeg",
            Arguments = ffmpegArgs,
            RedirectStandardOutput = false,
            RedirectStandardError = true,
            UseShellExecute = false,
            CreateNoWindow = true
        };

        using var process = Process.Start(processInfo);
        if (process == null)
        {
            Console.WriteLine("FFmpeg failed to start.");
            return;
        }

        var error = process.StandardError.ReadToEnd();

        process.WaitForExit();

        Console.WriteLine("FFmpeg Error: " + error);

        Console.WriteLine(File.Exists(combinedOutputFilePath)
            ? "Audio and video have been successfully combined."
            : "Failed to combine audio and video. Check the logs for errors.");
        File.Delete(videoFilePath);
        File.Delete(audioFilePath);
    }
}
